﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using ArcGIS.Core.CIM;
using ArcGIS.Core.Data;
using ArcGIS.Core.Geometry;
using ArcGIS.Desktop.Catalog;
using ArcGIS.Desktop.Core;
using ArcGIS.Desktop.Editing;
using ArcGIS.Desktop.Extensions;
using ArcGIS.Desktop.Framework;
using ArcGIS.Desktop.Framework.Contracts;
using ArcGIS.Desktop.Framework.Dialogs;
using ArcGIS.Desktop.Framework.Threading.Tasks;
using ArcGIS.Desktop.Layouts;
using ArcGIS.Desktop.Mapping;

namespace CIMSymbologyDemo.Tools
{
	internal class PlaceLineFeatureTool : MapTool
	{

		private FeatureLayer _lineFeatures = null;

		public PlaceLineFeatureTool()
		{
			IsSketchTool = true;
			SketchType = SketchGeometryType.Line;
			SketchOutputMode = SketchOutputMode.Map;
		}

		protected override Task OnToolActivateAsync(bool active)
		{
			_lineFeatures = MapView.Active.Map.GetLayersAsFlattenedList()
													 .FirstOrDefault(fl => fl.Name == "DemoLineFeatures") as FeatureLayer;
			return base.OnToolActivateAsync(active);
		}

		protected override Task<bool> OnSketchCompleteAsync(Geometry geometry)
		{
			if (_lineFeatures == null)
				return Task.FromResult(true);
			return QueuedTask.Run(() =>
			{
				var editOp = new EditOperation()
				{
					Name = "Create line feature",
					SelectNewFeatures = true,
					CancelMessage = "Cancelled"
				};

				editOp.Create(_lineFeatures, geometry);
				return editOp.Execute();
			});
		}
	}
}
